package ch.unige.solidify.controller;

import java.util.HashMap;
import java.util.Map;
import java.util.UUID;

import org.json.JSONObject;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.context.annotation.Profile;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.bind.annotation.RestController;
import org.springframework.web.reactive.function.BodyInserters;
import org.springframework.web.reactive.function.client.WebClient;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import reactor.core.publisher.Mono;

import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.config.SolidifyProperties;
import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.model.PersonWithOrcid;
import ch.unige.solidify.rest.ActionName;
import ch.unige.solidify.security.EveryonePermissions;
import ch.unige.solidify.security.UserPermissions;
import ch.unige.solidify.service.PersonWithOrcidService;

@RestController
@Profile("orcid")
@RequestMapping(SolidifyConstants.URL_ORCID)
public class OrcidController {
  private static final Logger log = LoggerFactory.getLogger(OrcidController.class);

  private static class PersonOrigin {
    private final String originUrl;
    private final String personId;

    public PersonOrigin(String personId, String originUrl) {
      this.personId = personId;
      this.originUrl = originUrl;
    }

    public String getOriginUrl() {
      return this.originUrl;
    }

    public String getPersonId() {
      return this.personId;
    }
  }

  private static final String GRANT_TYPE = "authorization_code";
  private static final String LOCATION_HEADER = "Location";
  private static final String ORCID_BASE_URL = "https://orcid.org/oauth";

  private static final String ORCID_TOKEN_URL = ORCID_BASE_URL + "/token";
  private final SolidifyProperties config;

  private final Map<String, PersonOrigin> mapPersonOrigin = new HashMap<>();

  private final PersonWithOrcidService personService;

  public OrcidController(PersonWithOrcidService personService, SolidifyProperties config) {
    this.personService = personService;
    this.config = config;
  }

  @EveryonePermissions
  @GetMapping(SolidifyConstants.URL_SEP + ActionName.LANDING)
  public ResponseEntity<Void> landingUrl(@RequestParam String code, @RequestParam String authTransactionId) throws JsonProcessingException {
    final PersonWithOrcid person = this.personService.findOne(this.mapPersonOrigin.get(authTransactionId).getPersonId());

    if (person == null) {
      throw new SolidifyRuntimeException("ORCID authentication : The received authTransactionId (" + authTransactionId + ") is not registered");
    }
    final WebClient client = WebClient.create(ORCID_TOKEN_URL);
    final Mono<String> response = client.post()
            .header("Accept", "application/json")
            .header("Content-Type", "application/x-www-form-urlencoded")
            .body(BodyInserters.fromValue("client_id=" + this.config.getOrcid().getClientId()
                    + "&client_secret=" + this.config.getOrcid().getClientSecret()
                    + "&grant_type=" + GRANT_TYPE
                    + "&code=" + code))
            .exchangeToMono(clientResponse -> clientResponse.bodyToMono(String.class));
    final JsonNode jsonResponse = new ObjectMapper().readTree(response.block());
    final String authenticatedOrcid = jsonResponse.get("orcid").textValue();
    person.setOrcid(authenticatedOrcid);
    person.setHasAuthenticatedOrcid(true);
    log.info("{} has obtained authenticated ORCID {}", person, authenticatedOrcid);
    this.personService.save(person);
    final String originUrl = this.mapPersonOrigin.get(authTransactionId).getOriginUrl();
    this.mapPersonOrigin.remove(authTransactionId);
    return ResponseEntity.status(HttpStatus.FOUND).header(LOCATION_HEADER, originUrl).build();
  }

  @UserPermissions
  @GetMapping(SolidifyConstants.URL_SEP + ActionName.START_ORCID_AUTH)
  public ResponseEntity<String> startOrcidAuth(@RequestParam String originUrl) {
    final PersonWithOrcid person = this.getCurrentPerson();
    final String authTransactionId = UUID.randomUUID().toString();
    this.mapPersonOrigin.put(authTransactionId, new PersonOrigin(person.getResId(), originUrl));
    log.info("{} starting ORCID authentication with authTransactionId {}", person, authTransactionId);
    final String jsonString = new JSONObject().put("authTransactionId", authTransactionId).toString();
    return ResponseEntity.ok().body(jsonString);
  }

  private PersonWithOrcid getCurrentPerson() {
    final Authentication authentication = SecurityContextHolder.getContext().getAuthentication();
    return this.personService.gePerson(authentication);
  }
}
