package ch.unige.solidify.index.indexing.inmemory;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.PageImpl;
import org.springframework.data.domain.Pageable;

import ch.unige.solidify.IndexDefinition;
import ch.unige.solidify.IndexDefinitionList;
import ch.unige.solidify.IndexProperties;
import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.index.indexing.IndexingService;
import ch.unige.solidify.index.settings.IndexingSettingsService;
import ch.unige.solidify.model.index.IndexMetadata;
import ch.unige.solidify.model.index.IndexSettings;

public abstract class InMemoryService<T extends IndexMetadata> extends IndexingService<T> {
  private static final Logger log = LoggerFactory.getLogger(InMemoryService.class);

  private final IndexingSettingsService<IndexSettings> settingsService;
  private final IndexData indexData;

  protected InMemoryService(
          IndexProperties indexProperties,
          IndexingSettingsService<IndexSettings> settingsService,
          IndexDefinitionList indexList,
          IndexData indexData) {
    super(indexProperties);
    this.settingsService = settingsService;
    this.indexData = indexData;
    // Create indexes
    for (final IndexSettings archiveIndex : this.settingsService.findAll(null)) {
      if (!this.indexData.getIndexes().containsKey(archiveIndex.getResId())) {
        final Map<String, T> metadata = new HashMap<>();
        this.indexData.getIndexes().put(archiveIndex.getResId(), (Map<String, IndexMetadata>) metadata);
      }
    }
    // Generate data
    if (indexProperties.isGenerateData()) {
      T archiveMetadata = this.getMetadataObject();
      archiveMetadata.setResId(SolidifyConstants.TEST_RES_ID);
      // Set metadata section
      final StringBuilder metadata = new StringBuilder();
      // @formatter:off
      metadata.append("{")
                .append("\"field\":\"value\",")
                .append("\"").append("aip-organizational-unit").append("\":\"").append(SolidifyConstants.TEST_RES_ID).append("\",")
                .append("\"").append("aip-access-level").append("\":\"").append("PUBLIC").append("\"")
              .append("}");
      // @formatter:on
      archiveMetadata.setMetadata(metadata.toString());
      // Save for each index
      for (IndexDefinition index : indexList.getIndexList()) {
        archiveMetadata.setIndex(index.getIndexName());
        this.save(index.getIndexName(), archiveMetadata);
        log.info("IndexMetadata create {}", archiveMetadata);
      }
    }
  }

  @Override
  public boolean delete(String indexName, T metadata) {
    this.checkIndex(metadata.getIndex());
    final Map<String, T> index = (Map<String, T>) this.indexData.getIndexes().get(metadata.getIndex());
    final String metadataId = metadata.getResId();
    if (index.containsKey(metadataId)) {
      index.remove(metadataId, metadata);
      return true;
    } else {
      return false;
    }
  }

  @Override
  public List<T> findAll(String indexName, T search) {
    this.checkIndex(indexName);
    return new ArrayList<>(((Map<String, T>) this.indexData.getIndexes().get(indexName)).values());
  }

  @Override
  public T findOne(String indexName, String id) {
    this.checkIndex(indexName);
    return ((Map<String, T>) this.indexData.getIndexes().get(indexName)).getOrDefault(id, null);
  }

  @Override
  public T save(String indexName, T metadata) {
    this.checkIndex(metadata.getIndex());
    final Map<String, T> index = (Map<String, T>) this.indexData.getIndexes().get(metadata.getIndex());
    final String metadataId = metadata.getResId();
    if (!index.containsKey(metadataId)) {
      index.put(metadataId, metadata);
      return metadata;
    } else {
      return null;
    }
  }

  @Override
  public Page<T> search(String indexName, String query, Pageable pageable) {
    this.checkIndex(indexName);
    final List<T> list = new ArrayList<>();
    final Map<String, T> metadataList = (Map<String, T>) this.indexData.getIndexes().get(indexName);
    metadataList.entrySet().stream()
            .filter(m -> m.getValue().getMetadata().toString().contains(query))
            .forEach(m -> list.add(m.getValue()));
    return new PageImpl<>(list, pageable, list.size());
  }

  @Override
  public T update(String indexName, T metadata) {
    this.checkIndex(metadata.getIndex());
    final Map<String, T> index = (Map<String, T>) this.indexData.getIndexes().get(metadata.getIndex());
    if (index.containsKey(metadata.getResId())) {
      index.replace(metadata.getResId(), metadata);
      return metadata;
    } else {
      return null;
    }
  }

  @Override
  protected void checkIndex(String indexName) {
    super.checkIndex(indexName);
    if (this.settingsService.findOne(indexName) == null) {
      final String msg = "Undefined index " + indexName;
      log.error(msg);
      throw new SolidifyRuntimeException(msg);
    }
  }

}
