package ch.unige.solidify.specification;

import java.util.ArrayList;
import java.util.List;
import javax.persistence.criteria.CriteriaBuilder;
import javax.persistence.criteria.CriteriaQuery;
import javax.persistence.criteria.Join;
import javax.persistence.criteria.Path;
import javax.persistence.criteria.Predicate;
import javax.persistence.criteria.Root;

import org.springframework.data.jpa.domain.Specification;

import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.rest.Resource;

public abstract class SolidifySpecification<T extends Resource> implements Specification<T> {

  protected final T criteria;

  protected final List<JoinedTableInValues> joinedTableInValues = new ArrayList<>();

  /**
   * Indicates if a criteria must be added for the root entity resId.
   * <p>
   * If 'true' the list returned when using the specification will only contain 1 or 0 element. It can be used to confirm that a specific entity
   * correspond to all given criteria: if the entity matches all of them it is returned in the list, otherwise en empty list is returned.
   */
  protected boolean filterOnRootResId = false;

  public SolidifySpecification(T criteria) {
    this.criteria = criteria;
  }

  public List<JoinedTableInValues> getJoinedTableInValues() {
    return this.joinedTableInValues;
  }

  public void addJoinTableInValues(JoinedTableInValues joinedTableValue) {
    this.joinedTableInValues.add(joinedTableValue);
  }

  public boolean getFilterOnRootResId() {
    return this.filterOnRootResId;
  }

  public void setFilterOnRootResId(boolean filterOnRootResId) {
    this.filterOnRootResId = filterOnRootResId;
  }

  @Override
  public Predicate toPredicate(Root<T> root, CriteriaQuery<?> query, CriteriaBuilder builder) {
    final List<Predicate> predicatesList = new ArrayList<>();

    this.completePredicatesList(root, query, builder, predicatesList);
    this.completeJoinedTableInValues(root, query, predicatesList);
    this.completeFilterOnRootResId(root, builder, predicatesList);

    return builder.and(predicatesList.toArray(new Predicate[predicatesList.size()]));
  }

  private void completeJoinedTableInValues(Root<T> root, CriteriaQuery<?> query, List<Predicate> predicatesList) {
    if (!this.joinedTableInValues.isEmpty()) {

      // Add criteria on eventual JOINed table values
      this.joinedTableInValues.forEach((joinedTableValue) -> {
        predicatesList.add(this.getPathFromExpression(root.join(joinedTableValue.getRelationName()), joinedTableValue.getFieldName())
                .in(joinedTableValue.getValues()));
      });

      // Set distinct to avoid duplicates if multiple JOIN or multiple matching values
      query.distinct(true);
    }
  }

  private void completeFilterOnRootResId(Root<T> root, CriteriaBuilder builder, List<Predicate> predicatesList) {
    if (this.filterOnRootResId) {
      predicatesList.add(builder.equal(root.get(SolidifyConstants.DB_RES_ID), this.criteria.getResId()));
    }
  }

  /**
   * Method allowing to get a Path or a sub Path by using paths names delimited by a colon (like in JPA path expression)
   * Examples of possible values for pathExpression:
   * - "fieldname"
   * - "embeddedId.fieldname"
   *
   * @param join
   * @param pathExpression
   * @return
   */
  private Path getPathFromExpression(Join join, String pathExpression) {
    String[] attributeNames = pathExpression.split("\\.");
    Path path = null;
    for (String attributeName : attributeNames) {
      if (path == null) {
        path = join.get(attributeName);
      } else {
        path = path.get(attributeName);
      }
    }
    return path;
  }

  protected abstract void completePredicatesList(Root<T> root, CriteriaQuery<?> query, CriteriaBuilder builder, List<Predicate> predicatesList);
}
