package ch.unige.solidify.service;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.StringReader;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import javax.xml.namespace.QName;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.boot.autoconfigure.condition.ConditionalOnBean;
import org.springframework.core.io.support.PathMatchingResourcePatternResolver;
import org.springframework.core.io.support.ResourcePatternResolver;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.PageRequest;
import org.springframework.data.domain.Pageable;
import org.springframework.data.domain.Sort;
import org.springframework.stereotype.Service;
import org.springframework.util.StreamUtils;

import ch.unige.solidify.OAIConstants;
import ch.unige.solidify.OAIProperties;
import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.business.OAISetService;
import ch.unige.solidify.controller.OAIController;
import ch.unige.solidify.exception.OAIException;
import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.model.oai.OAIRecord;
import ch.unige.solidify.model.oai.OAIRepositoryInfo;
import ch.unige.solidify.model.oai.OAISet;
import ch.unige.solidify.model.oai.OAIToken;
import ch.unige.solidify.model.xml.oai.v2.DeletedRecordType;
import ch.unige.solidify.model.xml.oai.v2.DescriptionType;
import ch.unige.solidify.model.xml.oai.v2.GetRecordType;
import ch.unige.solidify.model.xml.oai.v2.GranularityType;
import ch.unige.solidify.model.xml.oai.v2.HeaderType;
import ch.unige.solidify.model.xml.oai.v2.IdentifyType;
import ch.unige.solidify.model.xml.oai.v2.ListIdentifiersType;
import ch.unige.solidify.model.xml.oai.v2.ListMetadataFormatsType;
import ch.unige.solidify.model.xml.oai.v2.ListRecordsType;
import ch.unige.solidify.model.xml.oai.v2.ListSetsType;
import ch.unige.solidify.model.xml.oai.v2.MetadataFormatType;
import ch.unige.solidify.model.xml.oai.v2.MetadataType;
import ch.unige.solidify.model.xml.oai.v2.OAIPMHerrorType;
import ch.unige.solidify.model.xml.oai.v2.OAIPMHerrorcodeType;
import ch.unige.solidify.model.xml.oai.v2.OAIPMHtype;
import ch.unige.solidify.model.xml.oai.v2.RecordType;
import ch.unige.solidify.model.xml.oai.v2.RequestType;
import ch.unige.solidify.model.xml.oai.v2.SetType;
import ch.unige.solidify.model.xml.oai.v2.VerbType;
import ch.unige.solidify.model.xml.oai.v2.oai_dc.ElementType;
import ch.unige.solidify.model.xml.oai.v2.oai_dc.OaiDcType;
import ch.unige.solidify.model.xml.oai.v2.oai_dc.ObjectFactory;
import ch.unige.solidify.model.xml.oai.v2.oai_identifier.OaiIdentifierType;
import ch.unige.solidify.model.xml.xhtml.v5.Html;
import ch.unige.solidify.rest.RestCollection;
import ch.unige.solidify.rest.RestCollectionPage;
import ch.unige.solidify.specification.OAISetSpecification;
import ch.unige.solidify.util.CollectionTool;
import ch.unige.solidify.util.StringTool;
import ch.unige.solidify.util.XMLTool;

@Service
@ConditionalOnBean(OAIController.class)
public class OAIService {
  private static final Logger log = LoggerFactory.getLogger(OAIService.class);

  private JAXBContext jaxbContext;
  private List<Class<?>> xmlClasses = new ArrayList<>(Arrays.asList(OAIPMHtype.class, OaiDcType.class, OaiIdentifierType.class, Html.class));

  private final int oaiPmhPageSize;

  private final ObjectFactory oaiDcFactory = new ObjectFactory();
  // OAI Objects
  private final OAIRepositoryInfo oaiRepositoryInfo;
  private final OAISetService oaiSetService;
  // OAI Service
  private final OAISearchService searchService;
  private final OAIRecordService recordService;
  private final OAIMetadataService metadataService;

  private final Map<String, String> xlstList = new HashMap<>();

  public OAIService(
          OAIProperties oaiProperties,
          OAIRepositoryInfo oaiRepositoryInfo,
          OAISetService oaiSetService,
          OAISearchService searchService,
          OAIRecordService recordService,
          OAIMetadataService metadataService) {
    this.oaiRepositoryInfo = oaiRepositoryInfo;
    this.oaiPmhPageSize = oaiProperties.getPageSize();
    this.oaiSetService = oaiSetService;
    this.searchService = searchService;
    this.recordService = recordService;
    this.metadataService = metadataService;
    // JAXB contexts
    try {
      // Add project XML Classes
      this.xmlClasses.addAll(this.metadataService.getOaiXmlClasses());
      this.jaxbContext = JAXBContext.newInstance(this.xmlClasses.toArray(Class<?>[]::new));
      // Load XSLT files
      if (oaiProperties.isXslt()) {
        this.loadXsl();
      }
    } catch (JAXBException | IOException e) {
      throw new SolidifyRuntimeException(e.getMessage(), e);
    }
  }

  public OAIPMHerrorType getError(OAIPMHerrorcodeType code, String message) {
    final OAIPMHerrorType oaiError = new OAIPMHerrorType();
    oaiError.setCode(code);
    oaiError.setValue(message);
    return oaiError;
  }

  public IdentifyType getIdentify(String oaiUrl) {
    final IdentifyType oaiIdentify = new IdentifyType();
    oaiIdentify.setRepositoryName(this.oaiRepositoryInfo.getLongName());
    oaiIdentify.setBaseURL(oaiUrl);
    oaiIdentify.setProtocolVersion("2.0");
    oaiIdentify.getAdminEmail().add(this.oaiRepositoryInfo.getEmail());
    oaiIdentify.setEarliestDatestamp(this.oaiRepositoryInfo.getProductionDate());
    oaiIdentify.setDeletedRecord(DeletedRecordType.PERSISTENT);
    oaiIdentify.setGranularity(GranularityType.YYYY_MM_DD_THH_MM_SS_Z);
    // Add oai-identifier
    final DescriptionType identifierDesc = new DescriptionType();
    final OaiIdentifierType oaiIdentifier = new OaiIdentifierType();
    oaiIdentifier.setScheme(OAIConstants.OAI_PREFIX);
    oaiIdentifier.setRepositoryIdentifier(this.oaiRepositoryInfo.getDomain());
    oaiIdentifier.setDelimiter(OAIConstants.OAI_SEP);
    oaiIdentifier.setSampleIdentifier(
            OAIConstants.OAI_PREFIX
                    + OAIConstants.OAI_SEP
                    + this.oaiRepositoryInfo.getDomain()
                    + OAIConstants.OAI_SEP
                    + this.oaiRepositoryInfo.getPrefix()
                    + StringTool.generateResId());
    identifierDesc.setAny(new JAXBElement<>(new QName("http://www.openarchives.org/OAI/2.0/oai-identifier", "oai-identifier"),
            OaiIdentifierType.class, oaiIdentifier));
    oaiIdentify.getDescription().add(identifierDesc);
    // Add local description
    final DescriptionType repoSolution = new DescriptionType();
    repoSolution.setAny(this.oaiRepositoryInfo.getRepositoryDefinition());
    oaiIdentify.getDescription().add(repoSolution);
    return oaiIdentify;
  }

  public GetRecordType getRecord(String baseUrl, String identifier, String metadataPrefix) throws JAXBException {
    // Check Parameters
    this.checkParameters(identifier, metadataPrefix);

    // GetRecordType
    final OAIRecord oaiRecord = this.recordService.getRecord(this.getResIdFromOaiId(identifier));
    if (oaiRecord == null) {
      throw new OAIException(OAIPMHerrorcodeType.ID_DOES_NOT_EXIST, identifier);
    }
    final GetRecordType getRecord = new GetRecordType();
    getRecord.setRecord(this.getRecord(baseUrl, oaiRecord, metadataPrefix));
    return getRecord;
  }

  public RequestType getRequest(String url, String verb, String identifier, String metadataPrefix, String from, String until, String set,
          String token) {
    final RequestType req = new RequestType();
    req.setValue(url);
    if (!StringTool.isNullOrEmpty(verb)) {
      req.setVerb(VerbType.fromValue(verb));
    }
    if (!StringTool.isNullOrEmpty(identifier)) {
      req.setIdentifier(identifier);
    }
    if (!StringTool.isNullOrEmpty(metadataPrefix)) {
      req.setMetadataPrefix(metadataPrefix);
    }
    if (!StringTool.isNullOrEmpty(from)) {
      req.setFrom(from);
    }
    if (!StringTool.isNullOrEmpty(until)) {
      req.setUntil(until);
    }
    if (!StringTool.isNullOrEmpty(set)) {
      req.setSet(set);
    }
    if (!StringTool.isNullOrEmpty(token)) {
      req.setResumptionToken(token);
    }
    return req;
  }

  public VerbType getVerb(String verb) {
    return VerbType.fromValue(verb);
  }

  public VerbType[] getVerbs() {
    return VerbType.values();
  }

  public ListIdentifiersType listIdentifiers(String metadataPrefix, String from, String until, String set, String token) {
    // Check Parameters
    final OAIToken oaiToken = this.checkParameters(metadataPrefix, from, until, set, token);
    // Get Set Query
    String setQuery = null;
    if (oaiToken.getOaiParameters().getSet() != null) {
      setQuery = oaiToken.getOaiParameters().getSet().getQuery();
    }
    final RestCollection<OAIRecord> collection = this.searchService.searchOAIRecords(oaiToken.getOaiParameters().getFrom(),
            oaiToken.getOaiParameters().getUntil(), setQuery, oaiToken.getPageable());
    if (collection.get_page().getTotalItems() == 0) {
      throw new OAIException(OAIPMHerrorcodeType.NO_RECORDS_MATCH, "No identifier found");
    }
    final ListIdentifiersType idList = new ListIdentifiersType();
    final List<OAIRecord> list = CollectionTool.toList(collection.get_data());
    for (final OAIRecord oaiRecord : list) {
      idList.getHeader().add(this.getIdentifier(oaiRecord));
    }
    if (collection.get_page().hasNext()) {
      idList.setResumptionToken(oaiToken.getToken(collection.get_page().getTotalItems(), collection.get_page().getPreviousItems()));
    }
    return idList;
  }

  public ListMetadataFormatsType listMetadataFormats(String identifier) {
    if (identifier != null && !identifier.isEmpty() && this.recordService.getRecord(this.getResIdFromOaiId(identifier)) == null) {
      throw new OAIException(OAIPMHerrorcodeType.ID_DOES_NOT_EXIST, identifier);
    }

    final ListMetadataFormatsType mdFormats = new ListMetadataFormatsType();
    // OAI DC
    final MetadataFormatType oaiDc = new MetadataFormatType();
    oaiDc.setMetadataPrefix(OAIConstants.OAI_DC);
    oaiDc.setSchema("http://www.openarchives.org/OAI/2.0/oai_dc.xsd");
    oaiDc.setMetadataNamespace("http://www.openarchives.org/OAI/2.0/oai_dc");
    mdFormats.getMetadataFormat().add(oaiDc);
    // DataCite
    final MetadataFormatType dataCite = new MetadataFormatType();
    dataCite.setMetadataPrefix(OAIConstants.OAI_DATACITE);
    dataCite.setSchema("https://schema.datacite.org/meta/kernel-4.3/metadata.xsd");
    dataCite.setMetadataNamespace("http://datacite.org/schema/kernel-4");
    mdFormats.getMetadataFormat().add(dataCite);

    // schema.org
    final MetadataFormatType schemaOrg = new MetadataFormatType();
    schemaOrg.setMetadataPrefix(OAIConstants.OAI_SCHEMA_ORG);
    schemaOrg.setSchema("https://schema.org/Dataset");
    schemaOrg.setMetadataNamespace("https://schema.org/Dataset");
    mdFormats.getMetadataFormat().add(schemaOrg);
    return mdFormats;
  }

  public ListRecordsType listRecords(String baseUrl, String metadataPrefix, String from, String until, String set, String token)
          throws JAXBException {
    // Check Parameters
    final OAIToken oaiToken = this.checkParameters(metadataPrefix, from, until, set, token);
    String setQuery = null;
    if (oaiToken.getOaiParameters().getSet() != null) {
      setQuery = oaiToken.getOaiParameters().getSet().getQuery();
    }

    final RestCollection<OAIRecord> collection = this.searchService.searchOAIRecords(oaiToken.getOaiParameters().getFrom(),
            oaiToken.getOaiParameters().getUntil(), setQuery, oaiToken.getPageable());
    if (collection.get_page().getTotalItems() == 0) {
      throw new OAIException(OAIPMHerrorcodeType.NO_RECORDS_MATCH, "No record found");
    }
    final ListRecordsType recList = new ListRecordsType();
    final List<OAIRecord> list = CollectionTool.toList(collection.get_data());
    for (final OAIRecord oaiRecord : list) {
      recList.getRecord().add(this.getRecord(baseUrl, oaiRecord, oaiToken.getOaiParameters().getMetadataPrefix()));
    }
    if (collection.get_page().hasNext()) {
      recList.setResumptionToken(oaiToken.getToken(collection.get_page().getTotalItems(), collection.get_page().getPreviousItems()));
    }
    return recList;
  }

  public ListSetsType listSets(String token) {

    /*
     * Pagination may be given by an URL token
     */
    final OAIToken oaiToken = new OAIToken(this.oaiPmhPageSize, Sort.by("name"));
    if (!StringTool.isNullOrEmpty(token)) {
      oaiToken.setToken(token);
    }

    final Page<OAISet> oaiSetsPage = this.getOaiSetPage(oaiToken.getPageable());

    final ListSetsType setTypes = new ListSetsType();
    for (final OAISet oaiSet : oaiSetsPage.getContent()) {

      final SetType setType = new SetType();

      setType.setSetSpec(oaiSet.getSpec());
      setType.setSetName(oaiSet.getName());

      final OaiDcType oaiDc = this.oaiDcFactory.createOaiDcType();
      oaiDc.getTitleOrCreatorOrSubject().add(this.oaiDcFactory.createDescription(this.getOaiDcEntry(oaiSet.getDescription(), null)));
      final DescriptionType descriptionType = new DescriptionType();
      final QName qName = new QName("http://www.openarchives.org/OAI/2.0/oai_dc/", "dc");
      descriptionType.setAny(new JAXBElement<>(qName, OaiDcType.class, oaiDc));

      setType.getSetDescription().add(descriptionType);

      setTypes.getSet().add(setType);
    }

    return setTypes;
  }

  public String printOAIPMH(OAIPMHtype oai, String rootUrl) throws JAXBException {
    final Marshaller marshaller = this.metadataService.getMarshaller(this.jaxbContext);
    marshaller.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, Boolean.TRUE);
    if (!StringTool.isNullOrEmpty(rootUrl)) {
      marshaller.setProperty("com.sun.xml.bind.xmlHeaders", "<?xml-stylesheet type='text/xsl' href='" + rootUrl + "'?>");
    }
    marshaller.setProperty(Marshaller.JAXB_SCHEMA_LOCATION,
            "http://www.openarchives.org/OAI/2.0/ http://www.openarchives.org/OAI/2.0/OAI-PMH.xsd");
    final JAXBElement<OAIPMHtype> root = this.getXML(oai);
    final ByteArrayOutputStream bos = new ByteArrayOutputStream();
    marshaller.marshal(root, bos);
    return bos.toString();
  }

  private void addOaiSetInfo(HeaderType hdr, String resId) {
    final Pageable pageable = PageRequest.of(0, RestCollectionPage.DEFAULT_SIZE_PAGE);
    final Page<OAISet> oaiSetsPage = this.getOaiSetPage(pageable);
    for (final OAISet oaiSet : oaiSetsPage.getContent()) {
      if (this.searchService.isInSet(resId, oaiSet.getQuery(), pageable)) {
        hdr.getSetSpec().add(oaiSet.getSpec());
      }
    }
  }

  private void checkParameters(String identifier, String metadataPrefix) {
    if (StringTool.isNullOrEmpty(identifier)) {
      throw new OAIException(OAIPMHerrorcodeType.BAD_ARGUMENT, "Identifier is mandatory");
    }
    if (StringTool.isNullOrEmpty(metadataPrefix)) {
      throw new OAIException(OAIPMHerrorcodeType.BAD_ARGUMENT, "Metadata prefix is mandatory");
    }
  }

  private OAIToken checkParameters(String metadataPrefix, String from, String until, String set, String token) {
    final OAIToken oaiToken = new OAIToken(this.oaiPmhPageSize);

    // Token parameter
    if (!StringTool.isNullOrEmpty(token)) {
      oaiToken.setToken(token);
    } else {
      // Metadata Prefix parameter
      if (StringTool.isNullOrEmpty(metadataPrefix)) {
        throw new OAIException(OAIPMHerrorcodeType.BAD_ARGUMENT, "Metadata prefix is mandatory");
      }
      if (!metadataPrefix.equals(OAIConstants.OAI_DC) && !metadataPrefix.equals(OAIConstants.OAI_DATACITE) && !metadataPrefix
              .equals(OAIConstants.OAI_SCHEMA_ORG)) {
        throw new OAIException(OAIPMHerrorcodeType.CANNOT_DISSEMINATE_FORMAT, "Metadata prefix not supported");
      }
      oaiToken.getOaiParameters().setMetadataPrefix(metadataPrefix);

      // From parameter
      if (!StringTool.isNullOrEmpty(from)) {
        if (!StringTool.checkDate(from)) {
          throw new OAIException(OAIPMHerrorcodeType.BAD_ARGUMENT, "Wrong 'from' date format (" + from + ")");
        }
        oaiToken.getOaiParameters().setFrom(from);
      }
      // Until parameter
      if (!StringTool.isNullOrEmpty(until)) {
        if (!StringTool.checkDate(until)) {
          throw new OAIException(OAIPMHerrorcodeType.BAD_ARGUMENT, "Wrong 'until' date format (" + until + ")");
        }
        oaiToken.getOaiParameters().setUntil(until);
      }
      if (!StringTool.isNullOrEmpty(from) && !StringTool.isNullOrEmpty(until) && from.length() != until.length()) {
        throw new OAIException(OAIPMHerrorcodeType.BAD_ARGUMENT,
                "'from' and 'until' dates do not have the same granularity: " + from + " / " + until);
      }

      // Set parameter
      if (!StringTool.isNullOrEmpty(set)) {

        final OAISet oaiSet = this.oaiSetService.findBySpec(set);
        if (oaiSet != null) {
          oaiToken.getOaiParameters().setSet(oaiSet);
        } else {
          throw new OAIException(OAIPMHerrorcodeType.NO_SET_HIERARCHY, "Set " + set + " does not exist");
        }
      }
    }
    return oaiToken;
  }

  private String getArchiveUrl() {
    final String archiveUrl = this.oaiRepositoryInfo.getArchiveHomePage();
    if (StringTool.isNullOrEmpty(archiveUrl) || !archiveUrl.endsWith(SolidifyConstants.URL_SEP)) {
      throw new SolidifyRuntimeException("Missing 'archive-home-page' parameter");
    }
    return archiveUrl;
  }

  private ElementType getOaiDcEntry(String value, String language) {
    final ElementType fieldValue = new ElementType();
    fieldValue.setValue(value);
    if (!StringTool.isNullOrEmpty(language)) {
      fieldValue.setLang(language);
    }
    return fieldValue;
  }

  private HeaderType getIdentifier(OAIRecord oaiRecord) {
    final HeaderType hdr = new HeaderType();
    hdr.setIdentifier(this.getOaiId(oaiRecord.getResId()));
    hdr.setDatestamp(oaiRecord.getCreationDate().toString());
    this.addOaiSetInfo(hdr, oaiRecord.getResId());
    return hdr;
  }

  private String getOaiId(String resId) {
    return OAIConstants.OAI_PREFIX + OAIConstants.OAI_SEP + this.oaiRepositoryInfo.getDomain() + OAIConstants.OAI_SEP
            + this.oaiRepositoryInfo.getPrefix() + resId;
  }

  private String getResIdFromOaiId(String oaiId) {
    return oaiId.substring(oaiId.lastIndexOf(OAIConstants.OAI_SEP) + 1);
  }

  @SuppressWarnings("squid:S128")
  private Object getMetadata(String baseUrl, String metadataPrefix, OAIRecord oaiRecord) throws JAXBException {
    String xsl = OAIConstants.OAI_DATACITE + "2";
    final String archiveUrl = this.getArchiveUrl();
    switch (metadataPrefix) {
      case OAIConstants.OAI_DATACITE:
        return this.metadataService.getDataCiteMetadata(oaiRecord.getDataCiteMetadata());
      case OAIConstants.OAI_DC:
        xsl += OAIConstants.OAI_DC + ".xsl";
        if (this.xlstList.containsKey(xsl)) {
          return this.transformMetadata(oaiRecord.getDataCiteMetadata(), xsl, archiveUrl, oaiRecord.getResId());
        }
      case OAIConstants.OAI_SCHEMA_ORG:
        xsl += OAIConstants.OAI_SCHEMA_ORG + ".xsl";
        if (this.xlstList.containsKey(xsl)) {
          return this.transformMetadata(oaiRecord.getDataCiteMetadata(), xsl, archiveUrl, oaiRecord.getResId());
        }
      default:
        throw new OAIException(OAIPMHerrorcodeType.NO_METADATA_FORMATS, "No metadata for this record");
    }
  }

  private Page<OAISet> getOaiSetPage(Pageable pageable) {
    /*
     * filter on enabled OAI sets
     */
    final OAISet criteria = new OAISet();
    criteria.setEnabled(true);
    final OAISetSpecification spec = new OAISetSpecification(criteria);
    return this.oaiSetService.findAll(spec, pageable);
  }

  private RecordType getRecord(String baseUrl, OAIRecord oaiRecord, String metadataPrefix) throws JAXBException {
    final RecordType rec = new RecordType();
    // Header
    rec.setHeader(this.getIdentifier(oaiRecord));
    // Metadata
    final MetadataType md = new MetadataType();
    md.setAny(this.getMetadata(baseUrl, metadataPrefix, oaiRecord));
    rec.setMetadata(md);
    return rec;
  }

  private JAXBElement<OAIPMHtype> getXML(OAIPMHtype oai) {
    final QName qName = new QName("http://www.openarchives.org/OAI/2.0/", "OAI-PMH");
    return new JAXBElement<>(qName, OAIPMHtype.class, oai);
  }

  private void loadXsl() throws IOException {
    final ResourcePatternResolver resolver = new PathMatchingResourcePatternResolver(this.getClass().getClassLoader());
    for (final org.springframework.core.io.Resource xsl : resolver
            .getResources("classpath*:/" + OAIConstants.XSL_HOME + "/" + OAIConstants.OAI_DATACITE + "2*.xsl")) {
      log.info("Finding XML transformation: {}", xsl.getFilename());
      this.xlstList.put(xsl.getFilename(), StreamUtils.copyToString(xsl.getInputStream(), Charset.defaultCharset()));
    }

  }

  private Object transformMetadata(String datacite, String xslName, String archiveUrl, String resId) throws JAXBException {
    // XSL parameters
    final Map<String, String> params = new HashMap<>();
    params.put("archiveUrl", archiveUrl);
    params.put("resId", resId);
    // Apply XSL
    final StringReader newMd = new StringReader(XMLTool.transform(datacite, this.xlstList.get(xslName), params));
    return this.jaxbContext.createUnmarshaller().unmarshal(newMd);
  }

}
